//// INTRODUCTION ////
// Detection of DAPI-stained nuclei from fluorescent microscopy images.  
//
// This macro can also, optionally measure nuclear intensity in a separate green or red (EdU) image 
// and count the number of EdU positive and negative nuclei.  If using an EdU image, the final 
// character of the DAPI image name must be "b" (e.g. "testb.tif") and the EdU image must have exactly 
// the same name, but end in either "g" if green (e.g. "testg.tif") or "r" if red (e.g. "testr.tif").
//
// Date: 26th January 2021
// Authors: Stephen Cross and Kerry Wadey


//// PARAMETERS ////
minSize = 200; 			// Minimum 2D area of a nucleus (measured in px2 units)
maxSize = 10000;		// Maximum 2D area of a nucleus (measured in px2 units)
measureEdU = true; 		// When "true", a separate image will be used to measure EdU signal
channelEdU = 2			// (Only used when measureEdU=true) Channel containing EdU signal ("1" for red, "2" for green)
thresholdEdU = 30;		// (Only used when measureEdU=true) Threshold mean intensity for nuclei to be classed as EdU+



//// SETUP ////
// To ensure the macro runs the same each time, we need to set ImageJ to a specific "BlackBackground" state. 
setOption("BlackBackground", true);

// We can significantly speed up macro processing by entering batchmode.  At the end of the macro we
// will call a separate command to show the windows again.
setBatchMode("hide");


//// EXTRACTING DAPI CHANNEL ////
// Getting title of selected window (we will use this to access this window at a later date)
dapiRGBWin = getTitle();

// Displaying the current filename in the log window for reference
print("Processing file \""+dapiRGBWin+"\"");

// If the input image is in RGB format, we need to convert it to a "composite" image.  
// We can determine whether this is the case by checking the bit-depth of the image.  If
// RGB it will have a depth of 24 bits.
if (bitDepth() == 24) {
	// This separates the red, green and blue channels into individually-accessible channels.
	// MENU: Image > Color > Make Composite
	run("Make Composite");
	
	// Extracting the third channel from our composite image
	// MENU: Image > Duplicate...
	run("Duplicate...", "duplicate channels=3");

	// We no longer need the RGB image (which is now a composite), so we can select that window and close it.
	selectWindow(dapiRGBWin);
	close();
	
}

// Getting the title for the dapi channel and creating a duplicate.  We will need the original copy when 
// drawing outlines of detected nuclei later on.
dapiWinOrig = getTitle();
run("Duplicate...", "duplicate");
dapiWin = getTitle();



//// (OPTIONAL) EXTRACTING EDU CHANNEL ////
// If EdU intensity is to be measured, we need to select this window and extract the EdU channel.
// The following block of code will only run if "measureEdu" is set to "true".
if (measureEdU) {
	// We need to get the window with same name as the DAPI image, but ending in "g" or "r".
	// First, we need to replace the final "b" in the filename with "g" or "r".  We know the character 
	// immediately after "b" will be "." (for the file extension), so we can simply find and replace
	// "b." with "g." or "r.".
	if (channelEdU == 1) {
		eduRGBWin = replace(dapiRGBWin, "b.", "r.");
	} else if (channelEdU == 2) {
		eduRGBWin = replace(dapiRGBWin, "b.", "g.");
	}
		
	// Next, we need to repeat the same RGB to 8-bit processing steps as for the DAPI channel, but
	// this time extracting the channel corresponding to the EdU signal (i.e. red or green).
	selectWindow(eduRGBWin);
	if (bitDepth() == 24) {
		run("Make Composite");
		run("Duplicate...", "duplicate channels="+channelEdU);
		selectWindow(eduRGBWin);
		close();
	}
	
	eduWinOrig = getTitle();
	run("Duplicate...", "duplicate");
	eduWin = getTitle();

}



//// DETECTING NUCLEI ////
// Ensuring the DAPI image is active.
selectWindow(dapiWin);

// Using rolling ball background subtraction to remove long lengthscale noise due to out of focus light
run("Subtract Background...", "rolling=50");

// Reducing noise in the image using a 2D Gaussian filter with sigma=2px
run("Gaussian Blur...", "sigma=2");

// Binarising the image with the automatic Otsu threshold.  Following this, nuclei will be white (intensity = 255)
// on a black background (intensity = 0)
run("Auto Threshold", "method=Otsu white");

// Filling any small holes in the nuclei, where a "hole" is a fully-enclosed region of background-labelled
// pixels.  We need to get rid of these before running watershed splitting, else the nuclei may become 
// oversplit.
run("Fill Holes");

// Separating any adjacent nuclei, which have become fused as a single contiguous region during filtering and 
// binarisation.  This will create a single pixel-wide line between such adjacent objects.
run("Watershed");



//// SETTING MEASUREMENTS ////
// Before measuring the nuclei, we need to tell ImageJ what measurements it should take.  Since the measurements
// to be made depend on whether we're measuring EdU, we run different commands.
// MENU: Analyze > Set Measurements...
if (measureEdU) {
	// By default, ImageJ will take all measurements from the binary image showing the objects. For intensity 
	// measurements this means all values would be 0 (since it's a binary image).  We use the "redirect" option
	// to tell ImageJ to take intensity measurements from the corresponding pixels in the EdU image.
	run("Set Measurements...", "mean redirect="+eduWin+" decimal=4");
	
} else {
	// If simply counting nuclei we don't need any measurements enabled
	run("Set Measurements...", "redirect=None decimal=4");	
	
}


//// COUNTING NUCLEI ////
// Detecting contiguous regions of foreground-labelled pixels as nuclei.  Using the "size" parameter to control
// the accepted range of nucleus sizes (the minSize and maxSize variables are set at the top of the macro).
run("Analyze Particles...", "size="+minSize+"-"+maxSize+" show=Nothing add display clear");

// The number of nuclei detected is equal to the number of rows in the results table.  We can get this value
// using the "nResults" function.  The number of nuclei is then displayed in the log window using the "print"
// function.
nNuclei = nResults();
print("Number of nuclei = "+nNuclei);

// If measuring the EdU intensity, we can count the number of EdU positive and negative nuclei.
if (measureEdU) {
	// Creating a variable "nPos" which will store the current number of nuclei identified as being EdU positive.
	nPos = 0;

	// Using a for-loop to iterate over all detected nuclei.  The for-loop has the format:
	//
	// "for([starting condition];[continuation condition];[update]) {
	//		[code goes here]
	//  }
	//
	// The loop takes the following steps;
	// 1. Create the variable as defined in "starting condition" ("row" in this case with a value of "0")
	// 2. Test if the variable meets the "continuation condition" (here, that "row" is less than "nNuclei")
	// 3. Execute the block of code "code goes here"
	// 4. Updates the variable according to "update" ("row++" means add 1 to the current value of "row")
	// 5. Go back to step 2 and keep working through the steps until "continuation condition" is no longer met,
	//    at which point exit the loop and continue with the rest of the macro.
	for (row=0;row<nNuclei;row++) {		
		// Getting the intensity measurement for the current nucleus from the results table using the 
		// "getResult" function
		intensity = getResult("Mean", row);

		// If the intensity of this nucleus is above "thresholdEdU" (specified at the start of the macro), 
		// add 1 to "nPos".
		if (intensity > thresholdEdU) {
			nPos = nPos + 1;
		}
	}

	// The number of EdU negative nuclei is equal to the total number of nuclei, less the number of EdU 
	// positive nuclei.
	nNeg = nNuclei - nPos;

	// Displaying the number of EdU positive and negative nuclei in the log window
	print("Number of EdU+ nuclei = "+nPos);
	print("Number of EdU- nuclei = "+nNeg);
	
}



//// DISPLAYING DETECTED NUCLEI OUTLINES ////
// For validation that the macro is working correctly we want to see what was detected and (if relevant),
// see which were identified as EdU positive.  We can do this by drawing outlines of each nucleus on the input
// images.  These are generated differently depending on whether we have an EdU signal.
if (measureEdU) {
	// Combining the original DAPI and EdU images.  The outlines will be drawn on this image.
	run("Merge Channels...", "c1="+dapiWinOrig+" c3="+eduWinOrig+" create");
	
	// Using another for-loop to iterate over each nucleus.
	for (row=0;row<nNuclei;row++) {
		// The Analyze Particles step added each detected nucleus to ImageJ's ROI Manager.  We use the
		// following command to select the current nucleus from the ROI Manager.
		roiManager("Select", row);

		// As with the EdU positive counts, we get the current nucleus' EdU intensity.  If this is
		// greater than thresholdEdU, we draw a green outline, otherwise we draw it in blue.
		intensity = getResult("Mean", row);
		if (intensity > thresholdEdU) {
			// Adding the current ROI to the combined image as a green line.
			Overlay.addSelection("green");
		} else {
			// Adding the current ROI to the combined image as a blue line.
			Overlay.addSelection("white");
		}
	}	
} else {
	// If not separating nuclei based on EdU signal, we can just use the same outline colour for all nuclei.
	selectWindow(dapiWinOrig);

	// Using another for-loop to iterate over each nucleus.
	for (row=0;row<nNuclei;row++) {
		// Selecting the current nucleus in the ROI Manager and adding the outline (here, this is always blue).
		roiManager("Select", row);
		Overlay.addSelection("white");
	}
}

// Currently the outlines are only visible in ImageJ.  To stamp the outlines onto the image, we use the 
// "Flatten" command.
run("Flatten");



//// TIDYING WINDOWS ////
// During the macro many windows have been created.  Since we stored the names of these earlier on, we're
// able to select and close them one by one.
run("Select None");
selectWindow(dapiWin);
close();
if (measureEdU) {
	selectWindow(eduWin);
	close();
}

// Closing the results table showing intensity measurements
selectWindow("Results");
run("Close");

// Finally, we exit batch mode and display all windows
setBatchMode("exit and display");
